# Operator with there Precedence.
all_operators = {
    "+": {
        "precedence": 2,
        "argument": 2
    },
    "-": {
        "precedence": 2,
        "argument": 2
    },
    "*": {
        "precedence": 3,
        "argument": 2
    },
    "/": {
        "precedence": 3,
        "argument": 2
    },
    "%": {
        "precedence": 3,
        "argument": 2
    },
    "^": {
        "precedence": 4,
        "argument": 2
    },
    "abs": {
        "precedence": 4,
        "argument": 1
    },
    "ceil": {
        "precedence": 4,
        "argument": 1
    },
    "floor": {
        "precedence": 4,
        "argument": 1
    },
    "round": {
        "precedence": 4,
        "argument": 2
    },
    "factorial": {
        "precedence": 4,
        "argument": 1
    },
    "gcd": {
        "precedence": 4,
        "argument": 2
    },
    "exp": {
        "precedence": 4,
        "argument": 1
    },
    "pow": {
        "precedence": 4,
        "argument": 2
    },
    "sqrt": {
        "precedence": 4,
        "argument": 1
    },
    "log": {
        "precedence": 4,
        "argument": 1
    },
    "ln": {
        "precedence": 4,
        "argument": 1
    },
    "log2": {
        "precedence": 4,
        "argument": 1
    },
    "Log": {
        "precedence": 4,
        "argument": 2
    },
    "sin": {
        "precedence": 4,
        "argument": 1
    },
    "cos": {
        "precedence": 4,
        "argument": 1
    },
    "tan": {
        "precedence": 4,
        "argument": 1
    },
    "asin": {
        "precedence": 4,
        "argument": 1
    },
    "acos": {
        "precedence": 4,
        "argument": 1
    },
    "atan": {
        "precedence": 4,
        "argument": 1
    },
    "sinh": {
        "precedence": 4,
        "argument": 1
    },
    "cosh": {
        "precedence": 4,
        "argument": 1
    },
    "tanh": {
        "precedence": 4,
        "argument": 1
    },
    "asinh": {
        "precedence": 4,
        "argument": 1
    },
    "acosh": {
        "precedence": 4,
        "argument": 1
    },
    "atanh": {
        "precedence": 4,
        "argument": 1
    },
    "hypot": {
        "precedence": 4,
        "argument": 2
    },
    "deg": {
        "precedence": 4,
        "argument": 1
    },
    "rad": {
        "precedence": 4,
        "argument": 1
    },
    "?": {
        "precedence": 0.5,
        "argument": 2
    },
    ":": {
        "precedence": 0.5,
        "argument": 2
    },
    ">": {
        "precedence": 0.75,
        "argument": 2
    },
    "<": {
        "precedence": 0.75,
        "argument": 2
    },
    "|": {
        "precedence": 0.5,
        "argument": 2
    },
    "&": {
        "precedence": 0.5,
        "argument": 2
    },
    "=": {
        "precedence": 0.75,
        "argument": 2
    },
    "!": {
        "precedence": 0.75,
        "argument": 2
    },
    "~": {
        "precedence": 0.75,
        "argument": 1
    },
    "#": {
        "precedence": 0.75,
        "argument": 2
    },
    "@": {
        "precedence": 0.75,
        "argument": 2
    },
    "(": {
        "precedence": 0,
        "argument": 1
    },
    ")": {
        "precedence": 0,
        "argument": 1
    },
    ",": {
        "precedence": 1,
        "argument": 1
    }
}


from .handle_invalid_scenarios import ParserError
from .prescan import pre_scan
from .converter import convert
from .evaluator import evaluate
from .utils import *
class Parser:
    def prescan(self, exp):
        try:
            if isinstance(exp, dict):
                expr = exp.get("exp")
                exp = expr
            if exp is None:
                raise ParserError("Empty Expression")
            return pre_scan(exp)
        except ParserError as e:
            raise e

    def convert(self, exp, variables=None):
        try:
            if isinstance(exp, dict):
                expr = exp.get("exp")
                variables = exp.get("variables", None)
                exp = expr

            prescanned = self.prescan(exp)
            if variables is None:
                postfix = convert(prescanned, None)
            else:
                postfix = convert(prescanned, variables)
            return postfix
        except ParserError as e:
            raise e

    def evaluate(self, exp, convert=True, variables=None):
        try:
            if isinstance(exp, dict):
                expr = exp.get("exp")
                variables = exp.get("variables", None)
                convert = exp.get("convert", None)
                exp = expr
            if isinstance(convert, dict):
                variables = convert
                convert = True
            if convert is None or convert==True:
                postfix = self.convert(exp, variables)
            else:
                postfix = exp

            result = evaluate(postfix, variables)
            #print(result)
            if isinstance(result, bool):
                return result
            if is_int(result):
                return int(result)
            elif is_float(result):
                return float(result)
            else:
                return result
        except ParserError as e:
            raise e



def is_float(x):
    try:
        if isinstance(x, int) or isinstance(x, float) or isinstance(x, str):
            a = float(x)
        else:
            return False
    except ValueError:
        return False
    else:
        return True


def is_bool(x):
    try:
        if isinstance(x, bool) or isinstance(x, str) or isinstance(x, bool):
            a = bool(x)
        else:
            return False
    except ValueError:
        return False
    else:
        return True


def is_int(x):
    try:
        if isinstance(x, int) or isinstance(x, float) or isinstance(x, str):
            a = float(x)
            b = int(a)
        else:
            return False
    except ValueError:
        return False
    else:
        return a == b


def is_valid_operand(operand, variables):
    operand = operand.strip()
    return is_float(operand) or is_int(operand) or len(operand) == 0 or (
        variables is not None and operand in variables)



import setuptools
with open("README.md", "r") as fh:
    long_description = fh.read()
setuptools.setup(
     name='ganit',
     version='0.1.4',
     author="Palash Kanti Kundu",
     author_email="me@palash90.in",
     description="Ganit(गणित) means Calculation in Sanskrit. As the name suggests this is a calculation utility",
     long_description=long_description,
     long_description_content_type="text/markdown",
     url="https://github.com/Palash90/ganit/tree/master/python",
     packages=setuptools.find_packages(),
     classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
     ],
 )
# python3 setup.py sdist bdist_wheel
# python3 -m twine upload dist/*


from math import *
from .utils import *
from .handle_invalid_scenarios import handle_invalid_scenario
default_variables = {
    "pi": pi,
    "e": e,
    "tau": 2 * pi,
    "True": True,
    "False": False
}


def perform_operation(operator, arg):

    if operator == '+':
        return arg[1] + arg[0]
    elif operator == '-':
        return arg[1] - arg[0]
    elif operator == '*':
        return arg[1] * arg[0]
    elif operator == '/':
        return arg[1] / arg[0]
    elif operator == '%':
        return arg[1] % arg[0]
    elif operator == '^':
        return arg[1]**arg[0]
    elif operator == 'abs':
        return abs(arg[0])
    elif operator == 'ceil':
        return ceil(arg[0])
    elif operator == 'floor':
        return floor(arg[0])
    elif operator == 'round':
        if is_int(arg[0]):
            return round(arg[1], int(arg[0]))
        else:
            handle_invalid_scenario("Rounding takes only 'int' as second argument")
    elif operator == 'factorial':
        if is_int(arg[0]) == False or int(arg[0]) < 0:
            handle_invalid_scenario(
                "Value Error: " + format(arg[0]) +
                ". Factorial calculation can only be done for integers greater than 0.")
        return factorial(int(arg[0]))
    elif operator == 'gcd':
        if (is_int(arg[0]) == False) or (is_int(arg[1]) == False):
            handle_invalid_scenario(
                "Value Error: " + format(arg[0]) + " , " + format(arg[1]) +
                ". GCD can only be determined for integers.")
        return gcd(int(arg[1]), int(arg[0]))
    elif operator == 'exp':
        return exp(arg[0])
    elif operator == 'pow':
        return pow(arg[1], arg[0])
    elif operator == 'sqrt':
        if arg[0] < 0:
            handle_invalid_scenario("Square root only works for positive numbers")
        return sqrt(arg[0])
    elif operator == 'log':
        return log10(arg[0])
    elif operator == 'ln':
        return log(arg[0])
    elif operator == 'log2':
        return log2(arg[0])
    elif operator == 'Log':
        return log(arg[1], arg[0])
    elif operator == 'sin':
        return sin(arg[0])
    elif operator == 'cos':
        return cos(arg[0])
    elif operator == 'tan':
        return tan(arg[0])
    elif operator == 'asin':
        return asin(arg[0])
    elif operator == 'acos':
        return acos(arg[0])
    elif operator == 'atan':
        return atan(arg[0])
    elif operator == 'sinh':
        return sinh(arg[0])
    elif operator == 'cosh':
        return cosh(arg[0])
    elif operator == 'tanh':
        return tanh(arg[0])
    elif operator == 'asinh':
        return asinh(arg[0])
    elif operator == 'acosh':
        return acosh(arg[0])
    elif operator == 'atanh':
        return atanh(arg[0])
    elif operator == 'hypot':
        return hypot(arg[1], arg[0])
    elif operator == 'deg':
        return degrees(arg[0])
    elif operator == 'rad':
        return radians(arg[0])
    elif operator == '?':
        if isinstance(arg[1], bool) and arg[1] == True:
            return arg[0]
        elif isinstance(arg[1], bool) and arg[1] == False:
            return arg[1]
        else:
            handle_invalid_scenario("'?:' only works for boolean condition")
    elif operator == ":":
        if isinstance(arg[1], bool) and arg[1] == False:
            return arg[0]
        else:
            return arg[1]
    elif operator == ">":
        return arg[1] > arg[0]
    elif operator == "<":
        return arg[1] < arg[0]
    elif operator == "=":
        return arg[1] == arg[0]
    elif operator == "@":
        return arg[1] >= arg[0]
    elif operator == "#":
        return arg[1] <= arg[0]
    elif operator == '!':
        return arg[0] != arg[1]
    elif operator == '~':
        if not is_bool(arg[0]):
            handle_invalid_scenario("'!' only works for boolean")
        return not arg[0]
    elif operator == '|':
        if isinstance(arg[1], bool) and isinstance(arg[0], bool):
            return arg[0] or arg[1]
        else:
            handle_invalid_scenario("'|' only works for boolean conditions")
    elif operator == '&':
        if isinstance(arg[1], bool) and isinstance(arg[0], bool):
            return arg[0] and arg[1]
        else:
            handle_invalid_scenario("'&' only works for boolean conditions")
